/****************************************************************************
 * This file is part of the project AqDatabase.
 * AqDatabase (c) by 2019 Martin Preuss, all rights reserved.
 *
 * The license for this file can be found in the file COPYING which you
 * should have received along with this file.
 ****************************************************************************/



#ifndef AQDB_DB_H
#define AQDB_DB_H


#include <aqdatabase/aqdb.h>

#include <gwenhywfar/inherit.h>
#include <gwenhywfar/list1.h>
#include <gwenhywfar/list2.h>
#include <gwenhywfar/idlist64.h>
#include <gwenhywfar/url.h>
#include <gwenhywfar/stringlist.h>


#ifdef __cplusplus
extern "C" {
#endif


typedef struct AQDB_DB AQDB_DB;
GWEN_INHERIT_FUNCTION_LIB_DEFS(AQDB_DB, AQDB_API)
GWEN_LIST_FUNCTION_LIB_DEFS(AQDB_DB, AQDB_DB, AQDB_API)
GWEN_LIST2_FUNCTION_LIB_DEFS(AQDB_DB, AQDB_DB, AQDB_API)


#include <aqdatabase/aqdb_object.h>
#include <aqdatabase/aqdb_column.h>
#include <aqdatabase/aqdb_sessiondescr.h>




/** @name Database Functions
 *
 */
/*@{*/

/**
 * Create a database according to the given URL.
 * Currently the following backends are supported:
 * - "dir" : DbDir backend which stores the database in a given folder, creating one
 *           file per table
 * - "file": DbFile backend which uses a single structured file to store a database
 *
 * @return 0 if ok, error code otherwise
 * @param url URL specifying the backend to use and the file or folder to use
 *            (e.g. "file:/tmp/test.db")
 * @param pDb pointer to a variable receiving the database object created
 */
AQDB_API
int AQDB_DbFactory(const char *url, AQDB_DB **pDb);


/**
 * Increase the internal reference counter, so an additional call to
 * @ref AQDB_DB_free() is needed to actually free the database object.
 * @param db pointer to AQDB_DB object
 */
AQDB_API
void AQDB_DB_Attach(AQDB_DB *db);


/**
 * Release the ressources assigned to the given database object, including the object
 * itself.
 * @param db pointer to AQDB_DB object (NULL is alowed and will be ignored).
 */
AQDB_API
void AQDB_DB_free(AQDB_DB *db);


/**
 * Return the URL given to @ref AQDB_DbFactory().
 * @return URL given to @ref AQDB_DbFactory()
 * @param db pointer to AQDB_DB object
 */
AQDB_API
const GWEN_URL *AQDB_DB_GetUrl(const AQDB_DB *db);


/**
 * Set size of the internal object cache.
 *
 * @return 0 if ok, error code otherwise
 * @param db pointer to database object
 * @param maxEntries maximum number of entries to store in the cache
 * @param maxSize maximum size of the cache in bytes
 */
AQDB_API
int AQDB_DB_SetCache(AQDB_DB *db, uint32_t maxEntries, uint32_t maxSize);


/**
 * Open the database according to the given flags
 * (see @ref AQDB_ACTION_FLAGS_READ and others).
 * If it doesn't exist on disc an error code will be returned.
 * @param db pointer to AQDB_DB object
 * @param flags see @ref AQDB_ACTION_FLAGS_READ and others
 */
AQDB_API
int AQDB_DB_Open(AQDB_DB *db, uint32_t flags);


/**
 * Creates the database according to the given flags
 * (see @ref AQDB_ACTION_FLAGS_READ and others).
 * If it already exists on disc an error code will be returned.
 * The database is not open after this call.
 * @param db pointer to AQDB_DB object
 * @param flags see @ref AQDB_ACTION_FLAGS_READ and others
 */
AQDB_API
int AQDB_DB_Create(AQDB_DB *db, uint32_t flags);


/**
 * Close the database according to the given flags
 * (see @ref AQDB_ACTION_FLAGS_ABORT and others).
 * If it isn't open an error code will be returned.
 * @param db pointer to AQDB_DB object
 */
AQDB_API
int AQDB_DB_Close(AQDB_DB *db, uint32_t flags);


/**
 * Perform some consistency checks on the database and return its result.
 *
 * @return 0 if okay, error code otherwise
 * @param db pointer to AQDB_DB object
 */
AQDB_API
int AQDB_DB_Check(AQDB_DB *db, GWEN_BUFFER *resultTextBuf);

/*@}*/




/** @name Table Functions
 *
 */
/*@{*/

/**
 * Open a table within an already open database.
 * If the database isn't open or the table doesn't exist an error
 * code is returned.
 * @param db pointer to AQDB_DB object
 * @param tname table name (i.e. name of the type stored in the new table)
 * @param flags see @ref AQDB_ACTION_FLAGS_READ and others
 * @param pTableId ptr to a variable of type AQDB_ID to receive the table id to use for later access
 */
AQDB_API
int AQDB_DB_OpenTable(AQDB_DB *db,
                      const char *tname,
                      uint32_t flags,
                      AQDB_ID *pTableId);


/**
 * Closes an already open table within a database.
 * @param db pointer to AQDB_DB object
 * @param flags see @ref AQDB_ACTION_FLAGS_READ and others
 */
AQDB_API
int AQDB_DB_CloseTable(AQDB_DB *db,
                       AQDB_ID tableId,
                       uint32_t flags);


/**
 * Creates a table within an already open database.
 * It the database isn't open or the table already exists an error
 * code is returned.
 * @param db pointer to AQDB_DB object
 * @param tname table name (i.e. name of the type stored in the new table)
 * @param flags see @ref AQDB_ACTION_FLAGS_READ and others
 * @param columns list of @ref AQDB_COLUMN objects describing the columns of the table (i.e. the fields of the type)
 */
AQDB_API
int AQDB_DB_CreateTable(AQDB_DB *db,
                        const char *tname,
                        uint32_t flags,
                        const AQDB_COLUMN_LIST *columns);


/**
 * Return a list of table names for the given db.
 *
 * @param db pointer to AQDB_DB object
 * @param tableNameList list of strings each containing the name of a table stored in the database.
 */
AQDB_API
int AQDB_DB_GetTableNameList(AQDB_DB *db, GWEN_STRINGLIST *tableNameList);


/**
 * Return the list of columns for the given table.
 * The caller is responsible for freeing the object returned, if any.
 *
 * @return column list for the given table, NULL on error
 * @param db database
 * @param tableId id of the type table for which to get the column list (as returned by @ref AQDB_DB_OpenTable)
 */
AQDB_API
AQDB_COLUMN_LIST *AQDB_DB_GetColumnList(AQDB_DB *db, AQDB_ID tableId);


/*@}*/




/** @name Object Functions
 *
 */
/*@{*/



/**
 * Read an object from a database.
 *
 * @return 0 if ok, error code otherwise
 * @param db pointer to database object
 * @param tableId id of the table to read from (as returned by @ref AQDB_DB_OpenTable())
 * @param objectId id of the object to read (as returned by @ref AQDB_DB_QueryGetFirst()
 *                 and @ref AQDB_DB_QueryGetNext())
 * @param pObject pointer to a variable to receive the object read from the database
 */
AQDB_API
int AQDB_DB_ReadObject(AQDB_DB *db,
                       AQDB_ID tableId,
                       AQDB_ID objectId,
                       AQDB_OBJECT **pObject);


/**
 * Write an object to a database.
 *
 * The given object must have been read using @ref AQDB_DB_ReadObject.
 *
 * @return 0 if ok, error code otherwise
 * @param db pointer to database object
 * @param object pointer the object to write
 *
 */
AQDB_API
int AQDB_DB_WriteObject(AQDB_DB *db, const AQDB_OBJECT *object);


/**
 * Delete an object from a database.
 *
 * @return 0 if ok, error code otherwise
 * @param db pointer to database object
 * @param tableId id of the table to delete an object from (as returned by @ref AQDB_DB_OpenTable())
 * @param objectId id of the object to delete (as returned by @ref AQDB_DB_QueryGetFirst()
 *                 and @ref AQDB_DB_QueryGetNext())
 *
 */
AQDB_API
int AQDB_DB_DeleteObject(AQDB_DB *db,
                         AQDB_ID tableId,
                         AQDB_ID objectId);


/**
 * Add a new object to the database.
 *
 * @return 0 if ok, error code otherwise
 * @param db pointer to database object
 * @param tableId id of the table to add an object to (as returned by @ref AQDB_DB_OpenTable())
 * @param object new object to add to the database. Upon return the object's fields "ObjectId"
 *               and "TableId" are updated.
 */
AQDB_API
int AQDB_DB_AddObject(AQDB_DB *db,
                      AQDB_ID tableId,
                      AQDB_OBJECT *object);


/*@}*/




/** @name Query Functions
 *
 */
/*@{*/


/**
 * Submit a query for database objects. On success the id of the matching objects can be
 * retrieved via @ref AQDB_DB_QueryGetFirst() and @ref AQDB_DB_QueryGetNext().
 *
 * @param db pointer to database object
 * @param tableId id of the table to query objects from (as returned by @ref AQDB_DB_OpenTable())
 * @param refQueryId id of a reference query. If given, the results from the given query are used as input
 *                   for this new query. If 0 all objects of the selected table are inspected.
 * @param queryExpression expression to match objects from the database against
 * @param flags flags to control the query (see @ref AQDB_QUERY_FLAGS_STOPATFIRST)
 * @param pQueryId pointer to a variable to receive the id of the newly created query
 */
AQDB_API
int AQDB_DB_QuerySubmit(AQDB_DB *db,
                        AQDB_ID tableId,
                        AQDB_ID refQueryId,
                        const char *queryExpression,
                        uint32_t flags,
                        AQDB_ID *pQueryId);


/**
 * Get the first results to a query submit by @ref AQDB_DB_QuerySubmit().
 *
 * @return 0 if okay, error code otherwise (GWEN_ERROR_NOT_FOUND in case the list is empty)
 * @param db pointer to database object
 * @param queryId id of the query to read results from
 * @param pIdListPtr pointer to a buffer to receive ids
 * @param pIdListLen pointer to a variable which holds the available number of entries in the buffer,
 *                   upon return this variable is updated to reflect the number of entries stored
 */
AQDB_API
int AQDB_DB_QueryGetFirst(AQDB_DB *db,
                          AQDB_ID queryId,
                          AQDB_ID *pIdListPtr,
                          int *pIdListLen);


/**
 * Get the next results to a query submit by @ref AQDB_DB_QuerySubmit().
 *
 * @return 0 if okay, error code otherwise (GWEN_ERROR_NOT_FOUND in case the list is empty)
 * @param db pointer to database object
 * @param queryId id of the query to read results from
 * @param pIdListPtr pointer to a buffer to receive ids
 * @param pIdListLen pointer to a variable which holds the available number of entries in the buffer,
 *                   upon return this variable is updated to reflect the number of entries stored
 */
AQDB_API
int AQDB_DB_QueryGetNext(AQDB_DB *db,
                         AQDB_ID queryId,
                         AQDB_ID *pIdListPtr,
                         int *pIdListLen);


/**
 * Close a query submit via @ref AQDB_DB_QuerySubmit().
 *
 * @return 0 if okay, error code otherwise (GWEN_ERROR_NOT_FOUND in case the list is empty)
 * @param db pointer to database object
 * @param queryId id of the query to read results from
 */
AQDB_API
int AQDB_DB_QueryClose(AQDB_DB *db, AQDB_ID queryId);


/*@}*/




/** @name Session Management Functions
 *
 */
/*@{*/

/**
 * Enter edit mode.
 *
 * Changes to the database can only be made between sessions framed by calls to
 * AQDB_DB_BeginEdit() and @ref AQDB_DB_EndEdit().
 *
 * @return 0 if ok, error code otherwise
 * @param db pointer to database object
 * @param flags not used for now (the upper 16 bits are reserved for internal use by plugins).
 */
AQDB_API
int AQDB_DB_BeginEdit(AQDB_DB *db, uint32_t flags);

/**
 * Leave edit mode.
 *
 * Changes to the database can only be made between sessions framed by calls to
 * AQDB_DB_BeginEdit() and @ref AQDB_DB_EndEdit().
 *
 * @return 0 if ok, error code otherwise
 * @param db pointer to database object
 * @param flags use AQDB_ACTION_FLAGS_ABORT to abort an edit session
 */
AQDB_API
int AQDB_DB_EndEdit(AQDB_DB *db, uint32_t flags);


/**
 * Get a list of rollback sessions. Those can then be used to rollback the database
 * to a status before a given change.
 *
 * @return 0 if ok, error code otherwise
 * @param db pointer to database object
 * @param sl list to receive rollback sessions
 */
AQDB_API
int AQDB_DB_GetRollbackSessions(AQDB_DB *db, AQDB_SESSION_DESCR_LIST *sl);

/**
 * Undo all sessions down to the one with the given id.
 *
 * @return 0 if ok, error code otherwise
 * @param db pointer to database object
 * @param sid id of the session,  from which on all later sessions (including the one with the fiven id)
 *            will be rolled back
 */

AQDB_API
int AQDB_DB_UndoSession(AQDB_DB *db, AQDB_ID sid);


/*@}*/




/** @name Convenience Functions
 *
 */
/*@{*/

typedef void *GWENHYWFAR_CB(*AQDB_DB_FOR_EVERY_QUERY_OBJECT_FN)(AQDB_DB *db,
                                                                AQDB_ID tableId,
                                                                AQDB_OBJECT *o,
                                                                void *data);

/**
 * This function calls the given callback function for every object
 * which matches the given query expression until the callback function
 * returns non-NULL.
 * @param db pointer to database
 * @param tableId id of the type table which to query
 * @param refQueryId id of reference query (0 if none). This can be used to
 *   create subqueries.
 * @param queryExpression use NULL to match all objects
 * @param flags flags to control the query (see @ref AQDB_QUERY_FLAGS_STOPATFIRST)
 * @param fn callback function to call
 * @param data 4th argument to callback function
 * @param pResult pointer to a variable which receives the result returned by the
 *   last invocation of the callback function (e.g. the non-NULL value returned by that
 *   call).
 */
AQDB_API
int AQDB_DB_ForEveryQueryObject(AQDB_DB *db,
                                AQDB_ID tableId,
                                AQDB_ID refQueryId,
                                const char *queryExpression,
                                uint32_t flags,
                                AQDB_DB_FOR_EVERY_QUERY_OBJECT_FN fn,
                                void *data,
                                void **pResult);

/*@}*/





#ifdef __cplusplus
}
#endif


#endif

