/****************************************************************************
 * This file is part of the project AqDatabase.
 * AqDatabase (c) by 2007 Martin Preuss, all rights reserved.
 *
 * The license for this file can be found in the file COPYING which you
 * should have received along with this file.
 ****************************************************************************/


#ifndef AQDB_USERDB_H
#define AQDB_USERDB_H


#include <aqdatabase/aqdb.h>
#include <aqdatabase/aqdb_db.h>
#include <aqdatabase/user/aqdb_user.h>
#include <aqdatabase/user/aqdb_group.h>
#include <aqdatabase/user/aqdb_acl.h>



#define AQDB_PERMISSIONS_NONE    0x00000000
/** allow reading objects from table */
#define AQDB_PERMISSIONS_READ    0x00000001
#define AQDB_PERMISSIONS_MODIFY  0x00000002
/** allow writing objects from table*/
#define AQDB_PERMISSIONS_ADD     0x00000004
/** allow deleting objects from table */
#define AQDB_PERMISSIONS_DELETE  0x00000008
/** allow opening this table */
#define AQDB_PERMISSIONS_OPEN    0x00000010
/** Global permission to undo sessions */
#define AQDB_PERMISSIONS_UNDO    0x80000000



#ifdef __cplusplus
extern "C" {
#endif


/**
 * Add user and group management functions to any database.
 *
 * This modul does not encorce access conditions, it merely allows you to specify an access control environment
 * based on some entities described below.
 *
 * To enforce access control you can use the module introduced by @ref AQDB_Session_new().
 *
 * This module knows the following entrities.
 *
 * USERS
 *
 * Users are identified by a
 * - user id (AQDB_ID assigned by this module)
 * - user name (unique string assigned by the user, uniqueness is not enforced by this module, this must be done by the caller)
 * - credentials (a string containing credentials like e.g. password or public key, format not defined here)
 * - primary group (AQDB_ID of the group)
 * - secondary group (AQDB_ID of the group)
 *
 * GROUPS
 *
 * Groups are identified by a
 * - group id (AQDB_ID assigned by this module)
 * - group name (unique string assigned by the user, uniqueness is not enforced by this module, this must be done by the caller)
 *
 * ACL
 *
 * ACL represent access control entries. Each ACL object contains the following fields:
 * - subject id (AQDB_ID of the user or group for user ACL or group ACL respectively)
 * - table name (name of the type table for which this ACL applies)
 * - permissions (a uint32_t variable containing the permissions to apply, see @ref AQDB_PERMISSIONS_READ and following)
 *
 * This module creates the necessary hidden additional tables for the entities described above upon @ref AQDB_DB_Create() and
 * loads them upon @ref AQDB_DB_Open(). It also blocks callers from creating those tables manually, although reading and manipulating
 * those tables is not blocked to allow for user management via the standard DB interface.
 *
 * The user can then add users, groups and ACL entries using the following functions.
 *
 * This module also contains functions which help the AQDB_Session module to enforce the ACL restrictions.
 */
int AQDB_UserDb_Extend(AQDB_DB *db);
void AQDB_UserDb_Unextend(AQDB_DB *db);

AQDB_USER *AQDB_UserDb_ReadUserByName(AQDB_DB *db, const char *userName);
AQDB_USER *AQDB_UserDb_ReadUserById(AQDB_DB *db, AQDB_ID userId);
int AQDB_UserDb_UpdateUser(AQDB_DB *db, const AQDB_USER *user);

AQDB_GROUP *AQDB_UserDb_ReadGroupByName(AQDB_DB *db, const char *groupName);
AQDB_GROUP *AQDB_UserDb_ReadGroupById(AQDB_DB *db, AQDB_ID groupId);
int AQDB_UserDb_UpdateGroup(AQDB_DB *db, const AQDB_GROUP *group);

int AQDB_UserDb_AddGroup(AQDB_DB *db, AQDB_GROUP *group);
int AQDB_UserDb_AddUser(AQDB_DB *db, AQDB_USER *user);

int AQDB_UserDb_ReadUserPermissions(AQDB_DB *db, AQDB_ID userId, const char *tableName);
int AQDB_UserDb_ReadGroupPermissions(AQDB_DB *db, AQDB_ID groupId, const char *tableName);

AQDB_ACL_LIST *AQDB_UserDb_ReadAllUserAcl(AQDB_DB *db);
AQDB_ACL_LIST *AQDB_UserDb_ReadAllGroupAcl(AQDB_DB *db);

int AQDB_UserDb_AddUserAcl(AQDB_DB *db, AQDB_ID userId, const char *tableName, uint32_t permissions);
int AQDB_UserDb_AddGroupAcl(AQDB_DB *db, AQDB_ID groupId, const char *tableName, uint32_t permissions);

int AQDB_UserDb_DeleteUserAcl(AQDB_DB *db, AQDB_ID userId, const char *tableName);
int AQDB_UserDb_DeleteGroupAcl(AQDB_DB *db, AQDB_ID groupId, const char *tableName);


#ifdef __cplusplus
}
#endif



#endif

